/* MoreBalance includes */
#include "conf.h"
#include "sys.h"

/* System includes */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

/* Parser states */
#define STATE_START 0
#define STATE_PROTO 1
#define STATE_POSTPROTO 2
#define STATE_NODE 3
#define STATE_POSTNODE 4
#define STATE_PROTO1 5
#define STATE_POSTPROTO1 6
#define STATE_PROTO2 7
#define STATE_POSTPROTO2 8
#define STATE_TIME 9
#define STATE_TIMEUNIT 10
#define STATE_AS 11
#define STATE_MATCHING 12
#define STATE_MATCHING2 13
//#define STATE_PROTO3 14
#define STATE_BE 15
#define STATE_VERY 16
#define STATE_STOP 20

/* Give an detailed error message */
conf error(const char *buffer, int line, int position, const char *reason)
{
	conf c;
	int pos, num;

	c.valid = 0;
	c.verbose = 0;
	pos = 0;
	num = 0;

	fprintf(stderr, "ERROR: configuration error in line %i\n", line);
	fprintf(stderr, "%s\n", buffer);
	while((num < position) && (pos < strlen(buffer)))
	{
		if(buffer[pos] == ' ') num++;
		pos++;
		fprintf(stderr, "-");
	}
	fprintf(stderr, "^ here.\n");
	fprintf(stderr, "Possible reason: %s.\n", reason);

	return c;
}

/* Read the configuration from a file */
conf conf_read(const char *filename)
{
	conf c;
	FILE *f;
	char buf[1024];
	char *pretoken, *token, *tokenbuffer, *pretokenbuffer, *orig;
	int state;
	int port;
	int counter, lcounter, xcounter;
	long time;
	cluster *ctmp;
	qos *qtmp;
	int reverse;
	int istoken;

	c.valid = 0;
	c.verbose = 0;
	c.cluster = NULL;
	c.qos = NULL;
	c.qosnum = 0;
	c.clusternum = 0;

	lcounter = 0;

	f = fopen(filename, "r");
	if(!f) return c;

	/* Read every single line */
	while(fgets(buf, sizeof(buf), f))
	{
		lcounter++;

		buf[strlen(buf) - 1] = 0;
		if((!strlen(buf) || (buf[0] == '#'))) continue;
		orig = strdup(buf);
		pretoken = strtok_r(buf, "\"", &pretokenbuffer);
		counter = 0;
		xcounter = 0;
		state = STATE_START;
		reverse = 0;

		time = 0;
		qtmp = NULL;
		ctmp = NULL;

		/* Tokenize each line */
		while(pretoken)
		{
			if(orig[xcounter] != '\"')
			{
				token = strtok_r(pretoken, " ", &tokenbuffer);
				istoken = 0;
			}
			else
			{
				token = pretoken;
				istoken = 1;
			}
			while(token)
			{
				xcounter += strlen(token) + 1;
				if((state == STATE_START) || (state == STATE_PROTO) || (state == STATE_PROTO1) || state == (STATE_PROTO2))
				{
					if((state == STATE_START) && (!strcmp(token, "guarantee")))
					{
						qtmp = (qos*)malloc(sizeof(qos));
						state = STATE_PROTO1;
					}
					else if((state == STATE_START) && (!strcmp(token, "be")))
					{
						state = STATE_BE;
					}
					else
					{
						port = sys_port(token);
						if(port == 0) return error(orig, lcounter, counter, "wrong service");
						if((state == STATE_PROTO) || (state == STATE_START))
						{
							if(state == STATE_START)
							{
								ctmp = (cluster*)malloc(sizeof(cluster));
								ctmp->ports = NULL;
								ctmp->portopts = NULL;
								ctmp->nodes = NULL;
								ctmp->nodesnum = 0;
								ctmp->portsnum = 0;
								ctmp->destport = 0;
								ctmp->destportopts = 0;
								ctmp->pattern = NULL;
							}

							ctmp->ports = (int*)realloc(ctmp->ports, (ctmp->portsnum + 1) * sizeof(int));
							ctmp->ports[ctmp->portsnum] = port;
							ctmp->portopts = (int*)realloc(ctmp->portopts, (ctmp->portsnum + 1) * sizeof(int));
							ctmp->portopts[ctmp->portsnum] = sys_portoptions(token);
							ctmp->portsnum++;

							state = STATE_POSTPROTO;
						}
						else if(state == STATE_PROTO1)
						{
							qtmp->port1 = port;
							state = STATE_POSTPROTO1;
						}
						else if(state == STATE_PROTO2)
						{
							qtmp->port2 = port;

							if(reverse)
							{
								port = qtmp->port1;
								qtmp->port1 = qtmp->port2;
								qtmp->port2 = port;
							}

							state = STATE_POSTPROTO2;
						}
					}
				}
				else if((state == STATE_BE) || (state == STATE_VERY))
				{
					if(!strcmp(token, "verbose"))
					{
						c.verbose = 1;
						state = STATE_STOP;
					}
					else if((state == STATE_BE) && (!strcmp(token, "very")))
					{
						state = STATE_VERY;
					}
					else return error(orig, lcounter, counter, "unknown token; 'verbose' or 'very' expected");
				}
				else if(state == STATE_POSTPROTO)
				{
					if(!strcmp(token, "and")) state = STATE_PROTO;
					else if(!strcmp(token, "use")) state = STATE_NODE;
					else return error(orig, lcounter, counter, "unknown token; 'and' or 'use' expected");
				}
				else if(state == STATE_NODE)
				{
					ctmp->nodes = (char**)realloc(ctmp->nodes, (ctmp->nodesnum + 1 * sizeof(char*)));
					ctmp->nodes[ctmp->nodesnum] = strdup(token);
					ctmp->nodesnum++;

					state = STATE_POSTNODE;
				}
				else if(state == STATE_AS)
				{
					port = sys_port(token);
					if(port == 0) return error(orig, lcounter, counter, "wrong service");
					ctmp->destport = port;
					ctmp->destportopts = sys_portoptions(token);
				//state = STATE_PROTO3;
					state = STATE_POSTNODE;
				}
				else if(state == STATE_POSTNODE)
				{
					if(!strcmp(token, "and")) state = STATE_NODE;
					else if(!strcmp(token, "as")) state = STATE_AS;
					else if(!strcmp(token, "when")) state = STATE_MATCHING;
					else return error(orig, lcounter, counter, "unknown token; 'and' expected");
				}
				else if(state == STATE_POSTPROTO1)
				{
					if(!strcmp(token, "before"))
					{
						reverse = 1;

						state = STATE_PROTO2;
					}
					else if(!strcmp(token, "after")) state = STATE_PROTO2;
					else return error(orig, lcounter, counter, "unknown token; 'before' or 'after' expected");
				}
				else if(state == STATE_POSTPROTO2)
				{
					if(!strcmp(token, "for")) state = STATE_TIME;
					else return error(orig, lcounter, counter, "unknown token; 'for' expected");
				}
				else if(state == STATE_MATCHING)
				{
					if(!strcmp(token, "matching")) state = STATE_MATCHING2;
					else return error(orig, lcounter, counter, "unknown token; 'matching' expected");
				}
				else if(state == STATE_MATCHING2)
				{
					/*printf("*** match %s ***\n", token);*/
					ctmp->pattern = strdup(token);
					state = STATE_POSTNODE;
				}
				else if(state == STATE_TIME)
				{
					time = strtol(token, NULL, 10);
					if((time == LONG_MIN) || (time == LONG_MAX)) return error(orig, lcounter, counter, "number needed");
					state = STATE_TIMEUNIT;
				}
				else if(state == STATE_TIMEUNIT)
				{
					if(!strcmp(token, "seconds"))
						time = time;
					else if(!strcmp(token, "minutes"))
						time = time * 60;
					else if(!strcmp(token, "hours"))
						time = time * 3600;
					else return error(orig, lcounter, counter, "unkown time unit");
					qtmp->timeout = time;

					c.qos = (qos**)realloc(c.qos, (c.qosnum + 1) * sizeof(qos*));
					c.qos[c.qosnum] = qtmp;
					c.qosnum++;

					state = STATE_STOP;
				}
				else if(state == STATE_STOP)
				{
					return error(orig, lcounter, counter, "trailing invalid text");
				}
				else return error(orig, lcounter, counter, "unknown token");

				if(!istoken) token = strtok_r(NULL, " ", &tokenbuffer);
				else token = NULL;
				counter++;
			}

			pretoken = strtok_r(NULL, "\"", &pretokenbuffer);
			counter++;
		}

		/* Handle unfinished states */
		if(state != STATE_STOP)
		{
			if(state == STATE_POSTPROTO2)
			{
				qtmp->timeout = 60;

				c.qos = (qos**)realloc(c.qos, (c.qosnum + 1) * sizeof(qos*));
				c.qos[c.qosnum] = qtmp;
				c.qosnum++;
			}
			else if(state == STATE_POSTNODE)
			{
				c.cluster = (cluster**)realloc(c.cluster, (c.clusternum + 1) * sizeof(cluster*));
				c.cluster[c.clusternum] = ctmp;
				c.clusternum++;
			}
			else if(state == STATE_NODE) return error(orig, lcounter, counter, "node name expected");
			else if((state == STATE_PROTO) || (state == STATE_PROTO1) || (state == STATE_PROTO2))
			{
				return error(orig, lcounter, counter, "need to specify a service here");
			}
			else return error(orig, lcounter, counter, "missing end of configuration directive");
		}

		free(orig);
	}
	fclose(f);

	/* Return a valid configuration */
	c.valid = 1;

	return c;
}

/* Dump the whole configuration */
void conf_dump(conf c)
{
	int i, j, k;

	for(i = 0; i < c.qosnum; i++)
	{
		printf("* QoS: %i after %i [%li seconds]\n", c.qos[i]->port1, c.qos[i]->port2, c.qos[i]->timeout);
	}
	for(i = 0; i < c.clusternum; i++)
	{
		for(j = 0; j < c.cluster[i]->portsnum; j++)
		{
			printf("* Cluster: ");
			for(k = 0; k < c.cluster[i]->nodesnum; k++)
			{
				printf("%s ", c.cluster[i]->nodes[k]);
			}
			printf(" [port %i", c.cluster[i]->ports[j]);
			if(c.cluster[i]->portopts[j]) printf(" (tls)");
			printf("]");
			if(c.cluster[i]->destport)
			{
				printf(" [-> port %i", c.cluster[i]->destport);
				if(c.cluster[i]->destportopts) printf(" (tls)");
				printf("]");
			}
			printf("\n");
		}
	}
	fflush(NULL);
}

/* Optimize the configuration set */
void conf_optimize(conf *c)
{
	int i, j, k, l, x;
	cluster *ctmp;

	/* Erase duplicate entries in each cluster */
	for(i = 0; i < c->clusternum; i++)
	{
		for(k = 0; k < c->cluster[i]->portsnum; k++)
		{
			for(l = k + 1; l < c->cluster[i]->portsnum; l++)
			{
				if(c->cluster[i]->ports[k] == c->cluster[i]->ports[l])
				{
					/* Remove port from one entry */
					for(x = l + 1; x < c->cluster[i]->portsnum; x++)
						c->cluster[i]->ports[x - 1] = c->cluster[i]->ports[x];
					c->cluster[i]->portsnum--;
					c->cluster[i]->ports = (int*)realloc(c->cluster[i]->ports, (c->cluster[i]->portsnum * sizeof(int)));

					l--;
				}
			}
		}
	}

	/* Merge cluster entries with same ports into a new dedicated entry */
	for(i = 0; i < c->clusternum; i++)
	{
		for(j = i + 1; j < c->clusternum; j++)
		{
			for(k = 0; k < c->cluster[i]->portsnum; k++)
			{
				for(l = 0; l < c->cluster[j]->portsnum; l++)
				{
					if(c->cluster[i]->ports[k] == c->cluster[j]->ports[l])
					{
						/* Create new cluster */
						ctmp = (cluster*)malloc(sizeof(cluster));
						ctmp->portsnum = 1;
						ctmp->ports = (int*)malloc(sizeof(int));
						ctmp->ports[0] = c->cluster[i]->ports[k];
						ctmp->nodesnum = c->cluster[i]->nodesnum + c->cluster[j]->nodesnum;
						ctmp->nodes = (char**)malloc(ctmp->nodesnum * sizeof(char*));
						for(x = 0; x < c->cluster[i]->nodesnum; x++)
						{
							ctmp->nodes[x] = strdup(c->cluster[i]->nodes[x]);
						}
						for(x = 0; x < c->cluster[j]->nodesnum; x++)
						{
							ctmp->nodes[x + c->cluster[i]->nodesnum] = strdup(c->cluster[j]->nodes[x]);
						}
						ctmp->pattern = strdup(c->cluster[i]->pattern);
						ctmp->destport = c->cluster[i]->destport;
						ctmp->destportopts = c->cluster[i]->destportopts;
						ctmp->portopts = (int*)malloc(sizeof(int));
						ctmp->portopts[0] = c->cluster[i]->portopts[k];

						/* Insert into cluster list */
						c->clusternum++;
						c->cluster = (cluster**)realloc(c->cluster, c->clusternum * sizeof(cluster*));
						c->cluster[c->clusternum - 1] = ctmp;

						/* Remove ports from original two entries */
						for(x = k + 1; x < c->cluster[i]->portsnum; x++)
							c->cluster[i]->ports[x - 1] = c->cluster[i]->ports[x];
						c->cluster[i]->portsnum--;
						c->cluster[i]->ports = (int*)realloc(c->cluster[i]->ports, (c->cluster[i]->portsnum * sizeof(int)));
						for(x = l + 1; x < c->cluster[j]->portsnum; x++)
							c->cluster[j]->ports[x - 1] = c->cluster[j]->ports[x];
						c->cluster[j]->portsnum--;
						c->cluster[j]->ports = (int*)realloc(c->cluster[j]->ports, (c->cluster[j]->portsnum * sizeof(int)));

						k--;
						l--;
					}
				}
			}
		}
	}

	/* Delete empty clusters (no more ports registered) */
	for(i = 0; i < c->clusternum; i++)
	{
		if(c->cluster[i]->portsnum == 0)
		{
			for(j = 0; j < c->cluster[i]->nodesnum; j++)
			{
				free(c->cluster[i]->nodes[j]);
			}
			free(c->cluster[i]->nodes);
			free(c->cluster[i]->ports);
			free(c->cluster[i]->portopts);
			if(c->cluster[i]->pattern) free(c->cluster[i]->pattern);
			free(c->cluster[i]);

			for(j = i + 1; j < c->clusternum; j++)
			{
				c->cluster[j - 1] = c->cluster[j];
			}
			c->clusternum--;
			c->cluster = (cluster**)realloc(c->cluster, c->clusternum * sizeof(cluster*));

			i--;
		}
	}
}

