/* MoreBalance includes */
#include "conf.h"
#include "server.h"

/* System includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <getopt.h>

/* Global configuration */

static char *configfile = NULL;

/* Child function: startup MoreBalance */
static void launch()
{
	conf c;

	c = conf_read(configfile);

	if(c.valid)
	{
		printf("Configuration loaded.\n");
		fflush(NULL);

		conf_optimize(&c);
		conf_dump(c);

		printf("Start the server.\n");
		fflush(NULL);

		server_start(c);
	}
	else
	{
		fprintf(stderr, "Error reading configuration file.\n");
		exit(-1);
	}
}

/* Available options */
static struct option const options[] =
{
	{"configuration", required_argument, NULL, 'c'},
	{"help", no_argument, NULL, 'h'},
	{"version", no_argument, NULL, 'v'},
	{NULL, 0, NULL, 0}
};

/* Main function: main safety loop */
int main(int argc, char *argv[])
{
	pid_t pid;
	int status;
	int c;

	while((c = getopt_long(argc, argv, "c:hv", options, NULL)) != -1)
	{
		switch(c)
		{
			case 'c':
				configfile = optarg;
				break;
			case 'h':
				printf("MoreBalance - a sophisticated load balancing system\n");
				printf("See http://mindx.sourceforge.net/morebalance/\n");
				printf("\n");
				printf("Options:\n");
				printf("[-c | --configuration <file>] Use configuration from <file>\n");
				printf("\n");
				printf("[-h | --help                ] Show this help\n");
				printf("[-v | --version             ] Display version information\n");
				exit(0);
				break;
			case 'v':
				printf("MoreBalance 0.1+20021216\n");
				exit(0);
				break;
			default:
				printf("Unknown option, try --help.\n");
				exit(-1);
		}
	}

	if(!configfile) configfile = "morebalance.conf";

	printf(">> Startup.\n");
	fflush(NULL);

	while(1)
	{
		pid = fork();
		switch(pid)
		{
			case -1:
				fprintf(stderr, "Couldn't fork.\n");
				exit(-1);
				break;
			case 0:
				launch();
				exit(-1);
				break;
			default:
				break;
		}

		printf(">> Loop.\n");
		fflush(NULL);

		wait(&status);
		if(!WIFSIGNALED(status)) break;

		fprintf(stderr, ">> Child crashed. Restarting.\n");
	}

	return 0;
}

